%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% fix_rpo.pl  (03-09-02)
%  =rpo , >rpo,  >=rpo  amb la preced. i stat fixats fins ara
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% USES : list, precstat, a_intrpo
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% eq_rpo1(S,T,Prec,Stat)
% ge_rpo1(S,T,Prec,Stat)
% gr_rpo1(S,T,Prec,Stat)
% ig_or_ge_prec(F,G,Prec)
% check_stat(StatF,ArgsL,ArgsR,Prec,Stat,Op)
% aval_stat(StatF,ArgsL,ArgsR,F,Prec,Stat,mul,StatOut,Op)
% segur_no_iguals(S,T,Prec,Stat)
% iguals_possible(S,T,Prec,Stat)
% fixar_stat(ArgsL,ArgsR,Prec,Stat,StatOut,F,Op)
% simpl_args(StatF,Prec,Stat,ArgsS,ArgsT,DifST,DifTS)
% subterm_possible(LTerms,R,Prec,Stat)
% obtenir_possible_subt(LTerms,R,Term,Prec,Stat)
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


eq_rpo1(S,T,_,_):-S=T,!.
eq_rpo1(S,_,_,_):-is_variable(S),!,fail.
eq_rpo1(_,T,_,_):-is_variable(T),!,fail.
eq_rpo1(S,T,Prec,Stat):-
	S=..[F|ArgsS], T=..[G|ArgsT],
	igual(F,G,Prec),
	length21(ArgsS,N),!, length21(ArgsT,N),
	quin_stat(F,Stat,StatF),
	simpl_args(StatF,Prec,Stat,ArgsS,ArgsT,[],[]),!.
	

ge_rpo1(S,T,Prec,Stat):-
	eq_rpo1(S,T,Prec,Stat),!.
ge_rpo1(S,_,_,_):-is_variable(S),!,fail. % T no s igual a S i S es variable
ge_rpo1(S,T,Prec,Stat):-
	gr_rpo1(S,T,Prec,Stat),!.
ge_rpo1(S,T,Prec,Stat):-
	S=..[F|ArgsS],
	T=..[G|ArgsT],
	member22(F>=G,Prec),
	length21(ArgsS,N),!, length21(ArgsT,N),
	quin_stat(F,Stat,StatF),
	simpl_args(StatF,Prec,Stat,ArgsS,ArgsT,[],[]),!.


gr_rpo1(S,_,_,_):-is_variable(S),!,fail. 
gr_rpo1(S,T,Prec,Stat):-
	eq_rpo1(S,T,Prec,Stat),!,fail.
gr_rpo1(S,T,Prec,Stat):-
	member_term_igrpo(T,S,Prec,Stat),!.
gr_rpo1(_,T,_,_):-   
	is_variable(T),!,fail.
gr_rpo1(S,T,Prec,Stat):-
	S=..[F|_], T=..[G|ArgsT],
	member212(F>G,Prec),
	gr_rpo1_all(S,ArgsT,Prec,Stat),!.
gr_rpo1(S,T,Prec,Stat):-
	S=..[F|ArgsS], T=..[G|ArgsT],
	ig_or_ge_prec(F,G,Prec),
	quin_stat(F,Stat,StatF),
	simpl_args(StatF,Prec,Stat,ArgsS,ArgsT,DifST,DifTS),!,
	treat_args(DifST,DifTS,S,StatF,Prec,Stat),!.
gr_rpo1(S,T,Prec,Stat):-
	S=..[_|ArgsS],
	obtain_vars(T,VarsT),
	elim_subt_noT(ArgsS,VarsT,ArgsS1),!,
	del_repes(ArgsS1,ArgsS2),
	member22(Sub,ArgsS2),
	ge_rpo1(Sub,T,Prec,Stat),!.
	
ig_or_ge_prec(F,G,Prec):-igual(F,G,Prec),!.
ig_or_ge_prec(F,G,Prec):-member212(F>=G,Prec),!.

treat_args([],_,_,_,_,_):-!,fail.
treat_args(_,[],_,_,_,_):-!.
treat_args([A],L,_,_,Prec,Stat):-
	gr_rpo1_all(A,L,Prec,Stat).
treat_args([A|_],[X|L],S,lex,Prec,Stat):-
	gr_rpo1(A,X,Prec,Stat),!,
	gr_rpo1_all(S,L,Prec,Stat),!.
treat_args(ArgsS,ArgsT,S,lexr,Prec,Stat):-
	reverse1(ArgsS,[A|_]),
	reverse1(ArgsT,[X|L]),
	gr_rpo1(A,X,Prec,Stat),!,
	gr_rpo1_all(S,L,Prec,Stat),!.
treat_args(ArgsS,ArgsT,_,mul,Prec,Stat):-
% els iguals ja els hem tret, per tant segur que per tot subT
% cal un subtS>subtT
	sets_terms_disj_vars(ArgsS,ArgsT,Sets,'>',1), 
	multisetSets(Sets,Prec,Stat),!.

%%%%
gr_rpo1_all(_,[],_,_):-!.
gr_rpo1_all(S,[X|R],Prec,Stat):-
	gr_rpo1(S,X,Prec,Stat),!,
	gr_rpo1_all(S,R,Prec,Stat),!.

%%%%

elim_subt_noT([],_,[]):-!.
elim_subt_noT([S|R],Vars,[S|R1]):-
	contains(S,Vars),!,
	elim_subt_noT(R,Vars,R1),!.
elim_subt_noT([_|R],Vars,R1):-
	elim_subt_noT(R,Vars,R1),!.

%%%%

multisetSets([],_,_):-!.
multisetSets([[TermsL,TermsR]|Sets],Prec,Stat):-
	gr_multiset(TermsL,TermsR,Prec,Stat),!,
	multisetSets(Sets,Prec,Stat),!.


%atenci abans ja hem tret tots els iguals

gr_multiset(_,[],_,_):-!.
gr_multiset(TermsL,[R|TermsR],Prec,Stat):-
	member22(L,TermsL),
	gr_rpo1(L,R,Prec,Stat),!,
	gr_multiset(TermsL,TermsR,Prec,Stat),!.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

member_term_igrpo(_,[],_,_):-!,fail.
member_term_igrpo(X,[T|_],Prec,Stat):-member_term_igrpo(X,T,Prec,Stat),!.
member_term_igrpo(X,[_|R],Prec,Stat):-!,member_term_igrpo(X,R,Prec,Stat),!.
member_term_igrpo(S,T,Prec,Stat):-
	eq_rpo1(S,T,Prec,Stat),!.
member_term_igrpo(S,T,Prec,Stat):-
	T=..[_|Args],
	member_term_igrpo(S,Args,Prec,Stat).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

check_stat(mul,ArgsL,ArgsR,Prec,Stat,Op):-!,
	is_possible_mul(ArgsL,ArgsR,Prec,Stat,Op),!.
check_stat(lex,ArgsL,ArgsR,Prec,Stat,Op):-!,
	is_possible_lex(ArgsL,ArgsR,Prec,Stat,Op),!.
check_stat(lexr,ArgsL,ArgsR,Prec,Stat,Op):-!,
	is_possible_lexr(ArgsL,ArgsR,Prec,Stat,Op),!.
check_stat(cap,ArgsL,ArgsR,Prec,Stat,Op):-
	is_possible_mul(ArgsL,ArgsR,Prec,Stat,Op),!.
check_stat(cap,ArgsL,ArgsR,Prec,Stat,Op):-
	is_possible_lex(ArgsL,ArgsR,Prec,Stat,Op),!.
check_stat(cap,ArgsL,ArgsR,Prec,Stat,Op):-
	is_possible_lexr(ArgsL,ArgsR,Prec,Stat,Op),!.

aval_stat(cap,ArgsL,ArgsR,F,Prec,Stat,mul,StatOut,Op):-
	is_possible_mul(ArgsL,ArgsR,Prec,Stat,Op),
	\+(is_possible_lex(ArgsL,ArgsR,Prec,Stat,Op)),
	\+(is_possible_lexr(ArgsL,ArgsR,Prec,Stat,Op)),!,
	close_stat(Prec,Stat,StatOut,[F,mul]),!.
aval_stat(cap,ArgsL,ArgsR,F,Prec,Stat,lex,StatOut,Op):-
	is_possible_lex(ArgsL,ArgsR,Prec,Stat,Op),
	\+(is_possible_mul(ArgsL,ArgsR,Prec,Stat,Op)),
	\+(is_possible_lexr(ArgsL,ArgsR,Prec,Stat,Op)),!,
	close_stat(Prec,Stat,StatOut,[F,lex]),!.
aval_stat(cap,ArgsL,ArgsR,F,Prec,Stat,lexr,StatOut,Op):-
	is_possible_lexr(ArgsL,ArgsR,Prec,Stat,Op),
	\+(is_possible_lex(ArgsL,ArgsR,Prec,Stat,Op)),
	\+(is_possible_mul(ArgsL,ArgsR,Prec,Stat,Op)),!,
	close_stat(Prec,Stat,StatOut,[F,lexr]),!.
aval_stat(StatF,ArgsL,ArgsR,_,Prec,Stat,StatF,Stat,Op):-
	check_stat(StatF,ArgsL,ArgsR,Prec,Stat,Op),!.
	

%%
is_possible_mul(ArgsL,ArgsR,Prec,Stat,'='):-
	length21(ArgsL,N),!,length21(ArgsR,N),
	simpl_args(mul,Prec,Stat,ArgsL,ArgsR,LR,RL),
	check_simpl_args_mul(LR,RL,'=').
is_possible_mul(ArgsL,ArgsR,Prec,Stat,Op):-
	simpl_args(mul,Prec,Stat,ArgsL,ArgsR,LR,RL),
	check_simpl_args_mul(LR,RL,Op).

check_simpl_args_mul([],[],Op):-!, (Op='>=';Op='=').
check_simpl_args_mul([],_,_):-!,fail.
check_simpl_args_mul([_|_],[],Op):-!, (Op='>=';Op='>').
check_simpl_args_mul(ArgsL,ArgsR,'='):-!,
	sets_terms_disj_vars(ArgsL,ArgsR,Sets,'=',1),
	all_sets_ig_length(Sets).	
%totes son > o >=
check_simpl_args_mul(ArgsL,ArgsR,Op):-
	delete32(Arg,ArgsR,ArgsR1),
	obtain_vars(Arg,VarsA),!,
	subt_contains_vars(ArgsL,VarsA),!,
	check_simpl_args_mul(ArgsL,ArgsR1,Op),!.

subt_contains_vars(Terms,VarsA):-
	member22(X,Terms),
	contains(X,VarsA),!.

all_sets_ig_length([]):-!.
all_sets_ig_length([[L1,L2]|Sets]):-
	length21(L1,N),!,length21(L2,N),
	all_sets_ig_length(Sets),!.
		

%%
is_possible_lex(ArgsL,ArgsR,Prec,Stat,Op):-
	simpl_args(lex,Prec,Stat,ArgsL,ArgsR,LR,RL),
	check_simpl_args_lex(LR,RL,Prec,Stat,Op).

check_simpl_args_lex([],[],_,_,Op):-!, (Op='>=';Op='=').
check_simpl_args_lex([],_,_,_,_):-!,fail.
check_simpl_args_lex([_|_],[],_,_,Op):-!, (Op='>=';Op='>').
check_simpl_args_lex([X|_],[Y|_],Prec,Stat,'='):-
	segur_no_iguals(X,Y,Prec,Stat),!,fail.
check_simpl_args_lex([_|ArgsL],[_|ArgsR],Prec,Stat,'='):-!,
	check_simpl_args_lex(ArgsL,ArgsR,Prec,Stat,'='),!.
%a partir d'aqu > o >=
check_simpl_args_lex([X|ArgsL],[Y|ArgsR],Prec,Stat,Op):-
	obtain_vars(Y,VarsY),!,
	contains(X,VarsY),!,
	\+(gr_rpo1(Y,X,Prec,Stat)),
	(\+(eq_rpo1(X,Y,Prec,Stat));
	 is_possible_lex(ArgsL,ArgsR,Prec,Stat,Op)),!.

is_possible_lexr(ArgsL,ArgsR,Prec,Stat,Op):-
	reverse1(ArgsL,ArgsL1),
	reverse1(ArgsR,ArgsR1),
	is_possible_lex(ArgsL1,ArgsR1,Prec,Stat,Op),!.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

segur_no_iguals(S,T,_,_):-
	(is_variable(S);is_variable(T)),!,\+(S==T),!.
segur_no_iguals(S,T,Prec,Stat):-
	functor(S,K1,N1), 
	functor(T,K2,N2),
	(\+(N1=N2); segurnoigprec(K1,K2,Prec,Stat)),!.
segur_no_iguals(S,T,Prec,Stat):-		
	gr_rpo1(S,T,Prec,Stat),!.
segur_no_iguals(S,T,Prec,Stat):-		
	gr_rpo1(T,S,Prec,Stat),!.
segur_no_iguals(S,T,_,_):-
	obtain_vars(S,VarsS),
	obtain_vars(T,VarsT),!,
	difer2(VarsS,VarsT,VST,VTS),
	(VST=[_|_];VTS=[_|_]),!.

iguals_possible(S,T,Prec,Stat):-segur_no_iguals(S,T,Prec,Stat),!,fail.
iguals_possible(_,_,_,_):-!.

segurnoigprec(F,G,Prec,_):-member212(F>G,Prec),!.
segurnoigprec(F,G,Prec,_):-member212(G>F,Prec),!.
segurnoigprec(F,G,_,Stat):-
	quin_stat(F,Stat,StatF), \+(StatF=cap),
	quin_stat(G,Stat,StatG), \+(StatG=cap),
	\+(StatF=StatG),!.


%%%%%%%%%%%%%%%%%%%%%%%

%% comprova que Arg segur que no s igual segons rpo a cap
%% terme de la llista de termes (2n arg)
segur_no_igual_all(_,[],_,_):-!.
segur_no_igual_all(Arg,[Term|Rest],Prec,Stat):-
	segur_no_iguals(Arg,Term,Prec,Stat),!,
	segur_no_igual_all(Arg,Rest,Prec,Stat),!.


elim_gr_rpo1(_,[],_,_,[]):-!.
elim_gr_rpo1(Arg,[A|ArgsR],Prec,Stat,ArgsROut):-
	gr_rpo1(Arg,A,Prec,Stat),!,
	elim_gr_rpo1(Arg,ArgsR,Prec,Stat,ArgsROut),!.
elim_gr_rpo1(Arg,[A|ArgsR],Prec,Stat,[A|ArgsROut]):-
	elim_gr_rpo1(Arg,ArgsR,Prec,Stat,ArgsROut),!.
	
%%%%%%%%%%%%%%%%%%%%%%%%


fixar_stat(ArgsL,ArgsR,Prec,Stat,StatOut,F,Op):-
	is_possible_lex(ArgsL,ArgsR,Prec,Stat,Op),
	close_stat(Prec,Stat,StatOut,[F,lex]).
fixar_stat(ArgsL,ArgsR,Prec,Stat,StatOut,F,Op):-
	is_possible_lexr(ArgsL,ArgsR,Prec,Stat,Op),
	close_stat(Prec,Stat,StatOut,[F,lexr]).
fixar_stat(ArgsL,ArgsR,Prec,Stat,StatOut,F,Op):-
	is_possible_mul(ArgsL,ArgsR,Prec,Stat,Op),
	close_stat(Prec,Stat,StatOut,[F,mul]).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


simpl_args(mul,Prec,Stat,ArgsL,ArgsR,DifLR,DifRL):-
	difer_mul(Prec,Stat,ArgsL,ArgsR,DifLR,DifRL),!.
simpl_args(lex,Prec,Stat,ArgsL,ArgsR,DifLR,DifRL):-
	difer_lex(Prec,Stat,ArgsL,ArgsR,DifLR,DifRL),!.
simpl_args(lexr,Prec,Stat,ArgsL,ArgsR,DifLR,DifRL):-
	reverse1(ArgsL,ArgsL1),
	reverse1(ArgsR,ArgsR1),
	difer_lex(Prec,Stat,ArgsL1,ArgsR1,DifLR1,DifRL1),
	reverse1(DifLR1,DifLR),
	reverse1(DifRL1,DifRL),!.
simpl_args(cap,Prec,Stat,ArgsL,ArgsR,DifLR,DifRL):-
	length21(ArgsL,N),	%si no fixat noms es fa simpl. ms restrictiva
	length21(ArgsR,N),
	difer_lex(Prec,Stat,ArgsL,ArgsR,DifLR,DifRL),!.
simpl_args(cap,_,_,ArgsL,ArgsR,ArgsL,ArgsR):-!.



difer_mul(_,_,ArgsL,[],ArgsL,[]):-!.
difer_mul(_,_,[],ArgsR,[],ArgsR):-!.
difer_mul(Prec,Stat,[X|ArgsL],ArgsR,ArgsLR,ArgsRL):-
	delete33(Y,ArgsR,ArgsR1), 
	eq_rpo1(X,Y,Prec,Stat),!,
	difer_mul(Prec,Stat,ArgsL,ArgsR1,ArgsLR,ArgsRL),!.
difer_mul(Prec,Stat,[X|ArgsL],ArgsR,[X|ArgsLR],ArgsRL):-
	difer_mul(Prec,Stat,ArgsL,ArgsR,ArgsLR,ArgsRL),!.


% treiem els iguals en la mateixa posici.
% els que sn desprs del diferent s'eliminarien per subterme.
difer_lex(_,_,L,[],L,[]):-!.
difer_lex(_,_,[],L,[],L):-!.
difer_lex(Prec,Stat,[X|R],[Y|S],Dif1,Dif2):-
	eq_rpo1(X,Y,Prec,Stat),!,
	difer_lex(Prec,Stat,R,S,Dif1,Dif2),!.
difer_lex(Prec,Stat,[X|R],[Y|S],[X|Dif1],[Y|Dif2]):-
	difer_lex(Prec,Stat,R,S,Dif1,Dif2),!.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% comprovar si hi ha un terme de LTerms que pugui ser >=L segons
% les variables i no al revs rpo1 (fixat). (soluci SI/NO--> no back)

subterm_possible(LTerms,R,Prec,Stat):-
	obtain_vars(R,Rvars),!,
	member22(Term,LTerms),
	obtain_vars(Term,Lvars),
	difer(Rvars,Lvars,[]),
	\+(gr_rpo1(R,Term,Prec,Stat)),!.
	

%obtenir un terme de LTerms que pugui ser >=R segons les variables 
% i no gr_rpo1. Amb opci backtracking.

obtenir_possible_subt(LTerms,R,Term,Prec,Stat):-
	obtain_vars(R,Rvars),
	member22(Term,LTerms),
	obtain_vars(Term,Lvars),
	difer(Rvars,Lvars,[]),
	\+(gr_rpo1(R,Term,Prec,Stat)).
	

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

