%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% modul.pl   
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% aplica resultats de modularitat per unions disjuntes, 
% unions amb "constructor sharing" i unions jerrquiques
% (per unions jerarquiques cal us restringit de MSPO)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



mod([],['yes'],_,_,[]):-!.
mod(Rules,['mod',Resultat],H,Kind,Proofs):-
	inici(Rules),
	simplif_rules_strat(Rules,Rules1),
	leftR0(Def),
	retractall(leftRM(_)),
	assert(leftRM(Def)),
	obtain_systems(Rules1,Groups,Systems),
  % Groups=[[1,LGr1],[2,LGr2],...]  
  % Systems=[[LGrBD,GrDef],...] where 
  %	if LGrBD=[GrDef] then is a normal system
  %	otherwise LGrBD=[1,3,...] all basic systems and the hierarchical union
  %		  GrDef is a number in LGrBD which corresponds to the hierarchical union 
	ord_systems(Systems,Systems1),
  % First simple systems and then the hierarchical unions 
	prove_all_systems(Rules1,Groups,Systems1,Resultat,H,Kind,Proofs),!,
	retractall(leftRM(Def)).

obtain_systems(Rules,Groupsenum,Systems):-
	leftR0(Def), map(head,Def,Sdef),
	obtain_graph(Sdef,Rules,Graph),
  % Graph = [[f,Lf],[g,Lg],...] Lf is the list of all the defined symbols appearing in some
  %				rule headed defining f.
	obtain_groups(Sdef,Graph,Groups),
  % Groups = [G1,G2,...] GI is a grup of symbols defined at the same time
	gen_systems(Groups,Graph,Groupsenum,Systems),!.
  % Groupsenum=[[1,G1],[2,G2],...] with 
  %	GI=[f1,f2,...] f_i defined in the same system (usos creuats)

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
obtain_graph([],_,[]):-!.
obtain_graph([F|R],Rules,[[F,LF1]|Graph]):-
	obtain_rulesF_rest(Rules,F,RulesF,RestRules),
	noleftR0(Constructors), map(head,Constructors,Cons),
	get_all_symbols(RulesF,[],LF,Cons),
	delete32(F,LF,LF1),
	obtain_graph(R,RestRules,Graph).

obtain_rulesF_rest([],_,[],[]):-!.
obtain_rulesF_rest([[L,R]|Rules],F,[[L,R]|RulesF],RulesRest):-
	functor(L,F,_),!,
	obtain_rulesF_rest(Rules,F,RulesF,RulesRest),!.
obtain_rulesF_rest([Rule|Rules],F,RulesF,[Rule|RulesRest]):-
	obtain_rulesF_rest(Rules,F,RulesF,RulesRest),!.

get_all_symbols([],L,L,_):-!.
get_all_symbols([[L,R]|RulesF],LS0,LSymbols,Cons):-
	obtain_symbs_term(L,LS0,LS1),
	obtain_symbs_term(R,LS1,LS2),
	del_repes(LS2,LS3),
	difer(LS3,Cons,LS4),
	get_all_symbols(RulesF,LS4,LSymbols,Cons),!.
    
obtain_symbs_term(Term,Lsymb,Lsymb):-
	var(Term),!.
obtain_symbs_term(Term,Lsymb,LsymbOut):-
	functor(Term,F,_),
	Term=..[F|Args], 
	obtain_symbs_term_list(Args,[F|Lsymb],LsymbOut),!.

obtain_symbs_term_list([],Lsymb,Lsymb):-!.
obtain_symbs_term_list([Arg|Rest],Lsymb,LsymbOut):-
	obtain_symbs_term(Arg,Lsymb,Lsymb1),
	obtain_symbs_term_list(Rest,Lsymb1,LsymbOut),!.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

obtain_groups([],_,[]):-!.
obtain_groups([F|R],Graph,[GroupF|Rest]):-
	obtain_group(F,Graph,GroupF),!,
	difer(R,GroupF,R1),
	remove_g(GroupF,Graph,Graph1),!,
	obtain_groups(R1,Graph1,Rest),!.

remove_g([],Graph,Graph):-!.
remove_g([G|GroupF],Graph,GraphOut):-
	delete32([G,_],Graph,Graph1),
	remove_g(GroupF,Graph1,GraphOut),!.


obtain_group(F,Graph,GroupF):-
	obtain_routes([[F]],Graph,R),
	obtain_routes_group([F],R,GroupF).

obtain_routes([],_,[]):-!.
obtain_routes(LC,Graph,ROut):-
	findall(LR,routesTo(LC,Graph,LR),LRoutes),
	l_concat(LRoutes,LRoutes1),
	obtain_cyc_routes(LRoutes1,Routes,RestRoutes),
	obtain_routes(RestRoutes,Graph,RR),
	concat312(Routes,RR,ROut),!.

routesTo(L,Graph,LOut):-
	member22([F|R],L),
	findall([S,F|R],thereisway(S,F,Graph),LOut).

thereisway(S,F,Graph):-
	member22([S,L],Graph),
	member21(F,L).


obtain_cyc_routes([],[],[]):-!.
obtain_cyc_routes([[F|R]|LRoutes],[[F|R]|Routes],Rest):-
	member21(F,R),!,
	obtain_cyc_routes(LRoutes,Routes,Rest),!.
obtain_cyc_routes([C|LRoutes],Routes,[C|Rest]):-
	obtain_cyc_routes(LRoutes,Routes,Rest),!.



obtain_routes_group(LF,[],LF):-!.
obtain_routes_group(LF,Routes,GroupF):-
	sep_end_with(LF,Routes,Lgroup,Lrest),
	Lgroup=[_|_],!,
	l_union(Lgroup,LS),
	union(LS,LF,LF1),
	obtain_routes_group(LF1,Lrest,GroupF).
obtain_routes_group(LF,_,LF).

sep_end_with(_,[],[],[]):-!.
sep_end_with(LF,[[F|Fs]|R],[[F|Fs]|Lgroup],Lrest):-
	member21(F,LF),!,
	sep_end_with(LF,R,Lgroup,Lrest).
sep_end_with(LF,[[F|Fs]|R],Lgroup,[[F|Fs]|Lrest]):-
	sep_end_with(LF,R,Lgroup,Lrest).

	
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

gen_systems(Groups,Graph,Groupsenum,Systems):-
	enumerate(Groups,Groupsenum),
	length(Groups,N),
	crearlistIN(1,N,LN),
	cart_prod(LN,LN,[],Pairs),
	obtain_relations(Pairs,Groupsenum,Graph,RelGroups0),
	close_relations(RelGroups0,RelGroups),
	generate_systems(LN,RelGroups,Systems).

obtain_relations([],_,_,[]):-!.
obtain_relations([[N,M]|R],Groupsenum,Graph,[[N,M]|RelGroups]):-
	member22([N,GN],Groupsenum),
	member22([M,GM],Groupsenum),
	member22(F,GN),
	member21([F,L],Graph),
	inter(L,GM,[_|_]),!,
	delete032([M,N],R,R1),
	obtain_relations(R1,Groupsenum,Graph,RelGroups),!.
obtain_relations([_|R],Groupsenum,Graph,RelGroups):-
	obtain_relations(R,Groupsenum,Graph,RelGroups),!.


close_relations(RelGroups,ROut):-
	member22([N,M],RelGroups),
	member22([M,P],RelGroups),
	\+(member21([N,P],RelGroups)),!,
	close_relations([[N,P]|RelGroups],ROut),!.
close_relations(RelGroups,RelGroups):-!.



cart_prod([],_,P,P).
cart_prod([X|L1],L2,P1,P2):-
	cart(X,L2,P1,P3),
	cart_prod(L1,L2,P3,P2).

cart(_,[],P,P).
cart(X,[X|L],P1,P2):-!,
	cart(X,L,P1,P2).
cart(X,[Y|L],P1,[[X,Y]|P2]):-
	cart(X,L,P1,P2).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% retorna a Systems [[lgr,gi],....]
% on lgr=[gr1,gr2,...] s la llista dels grups de base
% i gi s el grup que representa la uni jerrquica
%
generate_systems([],_,[]):-!.
generate_systems([N|R],RelGroups,[[[N|LN],N]|Systems]):-
	findall(X,member22([N,X],RelGroups),LN),
	generate_systems(R,RelGroups,Systems),!.

	
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

ord_systems([],[]):-!.
ord_systems(Systems,SystemsOrd):-
	ord_systems_hier(Systems,[],SystemsOrd),!.

ord_systems_hier([],_,[]):-!.
ord_systems_hier(Systems,L,SystemsOrd):-
	selec_sys(Systems,L,Systems1,RestSystems),
	map(snd2,Systems1,L1),
	concat312(L,L1,L2),
	ord_systems_hier(RestSystems,L2,RestSystemsOrd),
	concat312(Systems1,RestSystemsOrd,SystemsOrd),!.

selec_sys([],_,[],[]):-!.
selec_sys([[LgrR,Gr]|Systems],L,[[LgrR,Gr]|BaseSystems],RestSystems):-
	difer(LgrR,[Gr|L],[]),!,
	selec_sys(Systems,L,BaseSystems,RestSystems),!.
selec_sys([S|Systems],L,BaseSystems,[S|RestSystems]):-
	selec_sys(Systems,L,BaseSystems,RestSystems),!.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

prove_all_systems(_,_,[],[],_,_,[]):-!.
prove_all_systems(Rules,Groups,[[LgrR,Gr]|Systems],[[Type|ResultSystem0]|Resultat],H,Kind,[Proof|Proofs]):-
	obtain_rules_systemM([Gr],Groups,Rules,LRulesHier,SymbsHier),
	((LgrR=[Gr],LRulesBase=LRulesHier,SymbsBase=SymbsHier,Type='n') % is a simple system
	;
	 (obtain_rules_systemM(LgrR,Groups,Rules,LRulesBase,SymbsBase), Type='m') % is a union
	),!,
	ordreM(LRulesBase,LRulesHier,SymbsBase,SymbsHier,Type,ResultSystem0,H,Kind,Proof),
	(
	 (ResultSystem0=['no'|_],
	  Resultat=[]
	 )
	 ;
	 (
	  prove_all_systems(Rules,Groups,Systems,Resultat,H,Kind,Proofs) 
 	 )
	),!.


obtain_rules_systemM(Lgroups,Groups,Rules,LRules,Lsymbols):-
	obtain_symbols_groups(Lgroups,Groups,[],Lsymbols),
	selec_rules_group(Lsymbols,Rules,LRules),!.

obtain_symbols_groups([],_,L,L):-!.
obtain_symbols_groups([G|Lgroups],Groups,L0,Lsymbols):-
	delete32([G,L],Groups,Groups1),
	concat312(L,L0,L1),!,
	obtain_symbols_groups(Lgroups,Groups1,L1,Lsymbols),!.	

selec_rules_group(_,[],[]):-!.
selec_rules_group(Lsymbols,[[L,R]|Rules],[[L,R]|LRules]):-
	functor(L,F,_), member21(F,Lsymbols),!,
	selec_rules_group(Lsymbols,Rules,LRules),!.
selec_rules_group(Lsymbols,[_|Rules],LRules):-
	selec_rules_group(Lsymbols,Rules,LRules),!.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Provar un sistema:

ordreM([],[],_,_,['yes'],_,_,[]):-!.
ordreM(LRules,_,Sbase,_,'n',['siMB',Resultat],H,Kind,[RulesEnum,CamiIni,ConstIniSimpl1,LproofSubgraphs]):- % simple system
	iniciM(LRules,Sbase),
	enumerate(LRules,RulesEnum),
	obt_const_trs(Kind,RulesEnum,LCAll),LCAll=[_|_],  % aix triv_or_cont: falla o s que si
	enumerate(LCAll,LCAllEnum),
	separ_cons(LCAllEnum,LCAllConstEnum),
  	triv_or_cont(LCAllConstEnum,LRules,Resultat,H,[N,LproofSubgraphs]),
	member21([N,[ConstIniSimpl1,CamiIni]],LCAllEnum),!.
ordreM(LRules,LRulesHier,Sbase,SHier,'m',['siMH',Sbase,SHier,Resultat],H,Kind,[RulesHierEnum,CamiIni,ConstIniSimpl1,LproofSubgraphs]):- %hierarchical union
	iniciM(LRules,Sbase),
   % to use modularity results for hierarchical unions we need to restrict
   % the kind of the constraints generated by MSPO to those which only have
   % defined symbols in the top of the constraints
	enumerate(LRulesHier,RulesHierEnum),
	const_trsM(RulesHierEnum,SHier,LCAll,Kind),LCAll=[_|_],  % triv_or_cont: falla o s que si
	enumerate(LCAll,LCAllEnum),
	separ_cons(LCAllEnum,LCAllConstEnum),
  	triv_or_cont(LCAllConstEnum,LRules,Resultat,H,[N,LproofSubgraphs]),!,
	member21([N,[ConstIniSimpl1,CamiIni]],LCAllEnum),!.
ordreM(_,_,Sbase,_,'n',['no',Sbase],_,_,[]):-!.
ordreM(_,_,Sbase,SHier,'m',['no',Sbase,SHier],_,_,[]):-!.


iniciM(LRules,Def):-
	variables(Vars),
	leftRM(Defined),
	selec_defined(Defined,Def,Def1),	
	clean,
	obtain_constructors(LRules,Def1,[],Cons),
	assertz(leftR(Def1)),
	assertz(noleftR(Cons)),
	assertz(variables(Vars)),
	assertz(leftR0(Def1)),
	assertz(noleftR0(Cons)),
	concat312(Def1,Cons,Symbs),
	limit_asserts(LRules,Symbs),!.


selec_defined([],_,[]):-!.
selec_defined([[D,Ar]|Defined],Def,[[D,Ar]|DefOut]):-
	member21(D,Def),!,
	selec_defined(Defined,Def,DefOut),!.
selec_defined([_|Defined],Def,DefOut):-
	selec_defined(Defined,Def,DefOut),!.
	

obtain_constructors([],_,Cons,Cons):-!.
obtain_constructors([[L,R]|LRules],S,C,ConsOut):-
	obtain_symbs(L,[],LS),
	obtain_symbs(R,LS,LS1),
	sort(LS1,LS2),
	difer(LS2,S,Cons),
	concat312(S,Cons,S1),
	concat312(C,Cons,C1),
	obtain_constructors(LRules,S1,C1,ConsOut),!.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%	
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%	
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%	
% Using modularity for "disjoint" and "constructor sharing" unions
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%	

% Kind is plain or twosteps or unaun or two_unaun

modBasic([],['yes'],_,_,[]):-!.
modBasic(Rules,Resultat,H,Kind,Proofs):-
	inici(Rules),
	simplif_rules_strat(Rules,Rules1),
	leftR0(Def),
	retractall(leftRM(_)),
	assert(leftRM(Def)),
	obtain_systems(Rules1,Groups,Systems),
	map(head, Systems,S1),
	obtain_systems_D_CS(S1,Systems1),
% Systems1=[[1,2,3],[4,5],...]
	one_or_more(Rules1,Groups,Systems1,Resultat,H,Kind,Proofs),!,
	retractall(leftRM(Def)).

one_or_more(Rules,_,[_],Resultat,H,Kind,Proof):-!, % un sol sistema
	ordre(Rules,Resultat,H,Kind,Proof),!.
one_or_more(Rules,Groups,Systems,['mod',Resultat],H,Kind,Proof):- 
			%mod indicar com mostrar el resultat
	prove_all_systems_D_CS(Rules,Groups,Systems,Resultat,H,Kind,Proof),!.


	
obtain_systems_D_CS([],[]):-!.
obtain_systems_D_CS([X|R],[Sx|R2]):-
	add_in_systems(X,R,R1,Sx),!,
	obtain_systems_D_CS(R1,R2),!.

add_in_systems(X,R,R1,Sx):-
	delete33(Y,R,R0),
	inter(X,Y,[_|_]),!,
	union(X,Y,Z),
	add_in_systems(Z,R0,R1,Sx),!.
add_in_systems(X,R,R,X).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Systems cont una llista de sistemes disjunts o constructor sharing.
% Per cada sistema provem l'ordre fins que tots provats o un falla.
% El resultat 'n' indica que no estem usant resultats de modularitat per unions jerarquiques
% Kind indica si usem plain(mspo00), twosteps(mspo), unaun(mspo1a1) or two_unaun(mspohard)

prove_all_systems_D_CS(_,_,[],[],_,_,[]):-!.
prove_all_systems_D_CS(Rules,Groups,[Lgr|Systems],[['n'|ResultSystem0]|Resultat],H,Kind,[Proof|Proofs]):-
	obtain_rules_systemM(Lgr,Groups,Rules,LRules,Symbs),!,
	ordre_D_CS(LRules,Symbs,ResultSystem0,H,Kind,Proof),
	(
	 ( ResultSystem0=['no'|_],
	   Resultat=[]
	 )
	 ;
	 (
	  prove_all_systems_D_CS(Rules,Groups,Systems,Resultat,H,Kind,Proofs) 
 	 )
	),!.


ordre_D_CS(LRules,Symbs,['siMB',Result],H,Kind,[LRulesEnum,CamiIni,ConstIniSimpl1,LproofSubgraphs]):-
	iniciM(LRules,Symbs),
	enumerate(LRules,LRulesEnum),
	obt_const_trs(Kind,LRulesEnum,LCAll), LCAll=[_|_],
	enumerate(LCAll,LCAllEnum),
	separ_cons(LCAllEnum,LCAllConstEnum),
	triv_or_cont(LCAllConstEnum,LRules,Result,H,[N,LproofSubgraphs]),
	member21([N,[ConstIniSimpl1,CamiIni]],LCAllEnum),!.
ordre_D_CS(_,Symbs,['no',Symbs],_,_,[]):-!.



